from __future__ import annotations

import os
from typing import TYPE_CHECKING
from unittest.mock import Mock
from uuid import uuid4

import pytest
import pytest_asyncio

from cashews import Cache
from cashews.backends.memory import Memory
from cashews.backends.transaction import TransactionBackend

if TYPE_CHECKING:  # pragma: no cover
    from cashews.backends.interface import Backend


pytest_plugins = ["pytest_asyncio"]  # pylint: disable=invalid-name

try:
    import aiohttp
except ImportError:
    pass
else:
    pytest_plugins.append("aiohttp.pytest_plugin")

    del aiohttp


@pytest.fixture
def redis_dsn():
    host = os.getenv("REDIS_HOST", "")
    port = os.getenv("REDIS_PORT", "6379")
    return f"redis://{host}:{port}/"


@pytest.fixture
def backend_factory():
    async def factory(backend_cls: type[Backend], *args, **kwargs):
        backend = backend_cls(*args, **kwargs)
        await backend.init()
        await backend.clear()
        return backend

    return factory


@pytest_asyncio.fixture(
    params=[
        "memory",
        "transactional",
        pytest.param("redis", marks=pytest.mark.redis),
        pytest.param("redis_hash", marks=pytest.mark.redis),
        pytest.param("redis_cs", marks=pytest.mark.redis),
        pytest.param("diskcache", marks=pytest.mark.diskcache),
    ],
)
async def backend(request, redis_dsn, backend_factory):
    if request.param == "diskcache":
        from cashews.backends.diskcache import DiskCache

        backend = await backend_factory(DiskCache, shards=0)
    elif request.param == "redis":
        from cashews.backends.redis import Redis

        backend = await backend_factory(
            Redis, redis_dsn, hash_key=None, max_connections=20, safe=False, socket_timeout=10
        )
    elif request.param == "redis_hash":
        from cashews.backends.redis import Redis

        backend = await backend_factory(
            Redis, redis_dsn, hash_key=uuid4().hex, max_connections=20, safe=False, socket_timeout=10
        )
    elif request.param == "redis_cs":
        from cashews.backends.redis.client_side import BcastClientSide

        backend = await backend_factory(
            BcastClientSide, redis_dsn, hash_key=None, max_connections=5, safe=False, socket_timeout=10
        )
    elif request.param == "transactional":
        backend = TransactionBackend(await backend_factory(Memory))
    else:
        backend = await backend_factory(Memory)
    try:
        yield backend
    finally:
        if request.param == "transactional":
            await backend.commit()
        await backend.close()


@pytest.fixture()
def target(backend):
    return Mock(wraps=backend, is_full_disable=False)


@pytest.fixture()
def cache(target):
    cache = Cache()
    cache._add_backend(target)
    return cache



import uuid
from abc import ABCMeta, abstractmethod
from contextlib import asynccontextmanager
from contextvars import ContextVar
from typing import Any, AsyncIterator, Mapping, Optional, Set, Tuple

from cashews.commands import ALL, Command
from cashews.exceptions import LockedError

NOT_EXIST = -2
UNLIMITED = -1


class _BackendInterface(metaclass=ABCMeta):
    @property
    @abstractmethod
    def is_init(self) -> bool:
        ...

    @abstractmethod
    async def init(self):
        ...

    @abstractmethod
    async def close(self):
        ...

    @abstractmethod
    async def set(
        self,
        key: str,
        value: Any,
        expire: Optional[float] = None,
        exist: Optional[bool] = None,
    ) -> bool:
        ...

    @abstractmethod
    async def set_raw(self, key: str, value: Any, **kwargs: Any):
        ...

    @abstractmethod
    async def get(self, key: str, default: Optional[Any] = None) -> Any:
        ...

    @abstractmethod
    async def get_raw(self, key: str) -> Any:
        ...

    @abstractmethod
    async def get_many(self, *keys: str, default: Optional[Any] = None) -> Tuple[Optional[Any], ...]:
        ...

    @abstractmethod
    async def set_many(self, pairs: Mapping[str, Any], expire: Optional[float] = None):
        ...

    @abstractmethod
    async def exists(self, key: str) -> bool:
        ...

    @abstractmethod
    async def scan(self, pattern: str, batch_size: int = 100) -> AsyncIterator[str]:
        ...

    @abstractmethod
    async def incr(self, key: str) -> int:
        ...

    @abstractmethod
    async def delete(self, key: str) -> bool:
        ...

    @abstractmethod
    async def delete_many(self, *keys: str):
        ...

    @abstractmethod
    async def delete_match(self, pattern: str):
        ...

    @abstractmethod
    async def get_match(self, pattern: str, batch_size: int = 100) -> AsyncIterator[Tuple[str, Any]]:
        ...

    @abstractmethod
    async def expire(self, key: str, timeout: float):
        ...

    @abstractmethod
    async def get_expire(self, key: str) -> int:
        ...

    @abstractmethod
    async def get_bits(self, key: str, *indexes: int, size: int = 1) -> Tuple[int, ...]:
        ...

    @abstractmethod
    async def incr_bits(self, key: str, *indexes: int, size: int = 1, by: int = 1) -> Tuple[int, ...]:
        ...

    @abstractmethod
    async def slice_incr(self, key: str, start: int, end: int, maxvalue: int, expire: Optional[float] = None) -> int:
        ...

    @abstractmethod
    async def get_size(self, key: str) -> int:
        """
        Return size in bites that allocated by a value for given key
        """
        ...

    @abstractmethod
    async def ping(self, message: Optional[bytes] = None) -> bytes:
        ...

    @abstractmethod
    async def clear(self):
        ...

    @abstractmethod
    async def set_lock(self, key: str, value: Any, expire: float) -> bool:
        ...

    @abstractmethod
    async def is_locked(
        self,
        key: str,
        wait: Optional[float] = None,
        step: float = 0.1,
    ) -> bool:
        ...

    @abstractmethod
    async def unlock(self, key: str, value: Any) -> bool:
        ...

    @asynccontextmanager
    async def lock(self, key: str, expire: float):
        identifier = str(uuid.uuid4())
        lock = await self.set_lock(key, identifier, expire=expire)
        if not lock:
            # we need to check the connection by ping because
            # if for example a redis unavailable and a backend have flag `safe`
            # we will have a brake lock
            try:
                await self.ping(b"TEST")
            except Exception:
                pass
            else:
                raise LockedError(f"Key {key} already locked")
        try:
            yield
        finally:
            await self.unlock(key, identifier)


class ControlMixin:
    def __init__(self) -> None:
        self.__disable: ContextVar[Set[Command]] = ContextVar(str(id(self)), default=set())

    @property
    def _disable(self) -> Set[Command]:
        return self.__disable.get()

    def _set_disable(self, value: Set[Command]) -> None:
        self.__disable.set(value)

    def is_disable(self, *cmds: Command) -> bool:
        _disable = self._disable
        if not cmds and _disable:
            return True
        for cmd in cmds:
            if cmd in _disable:
                return True
        return False

    def is_enable(self, *cmds: Command) -> bool:
        return not self.is_disable(*cmds)

    @property
    def is_full_disable(self):
        return self._disable == ALL

    def disable(self, *cmds: Command) -> None:
        if not cmds:
            _disable = ALL.copy()
        else:
            _disable = self._disable.copy()
            _disable.update(cmds)
        self._set_disable(_disable)

    def enable(self, *cmds: Command) -> None:
        if not cmds:
            _disable = set()
        else:
            _disable = self._disable.copy()
            _disable -= set(cmds)
        self._set_disable(_disable)


class Backend(ControlMixin, _BackendInterface, metaclass=ABCMeta):
    def __init__(self, *args, **kwargs):
        super().__init__()



from functools import wraps
from typing import Optional

from cashews._typing import TTL, AsyncCallable_T, CallableCacheCondition, Decorator
from cashews.backends.interface import _BackendInterface
from cashews.formatter import register_template
from cashews.key import get_cache_key, get_cache_key_template
from cashews.ttl import ttl_to_seconds

from .defaults import _empty, context_cache_detect

__all__ = ("cache",)


def cache(
    backend: _BackendInterface,
    ttl: TTL,
    key: Optional[str] = None,
    condition: CallableCacheCondition = lambda *args, **kwargs: True,
    prefix: str = "",
) -> Decorator:
    """
    Simple cache strategy - trying to return cached result,
    execute wrapped call and store a result with ttl if condition return true
    :param backend: cache backend
    :param ttl: duration in seconds to store a result or a callable
    :param key: custom cache key, may contain alias to args or kwargs passed to a call
    :param condition: callable object that determines whether the result will be saved or not
    :param prefix: custom prefix for key
    """

    ttl = ttl_to_seconds(ttl)

    def _decor(func: AsyncCallable_T) -> AsyncCallable_T:
        _key_template = get_cache_key_template(func, key=key, prefix=prefix)
        register_template(func, _key_template)

        @wraps(func)
        async def _wrap(*args, **kwargs):
            _ttl = ttl_to_seconds(ttl, *args, **kwargs, with_callable=True)
            _cache_key = get_cache_key(func, _key_template, args, kwargs)
            cached = await backend.get(_cache_key, default=_empty)
            if cached is not _empty:
                context_cache_detect._set(_cache_key, ttl=_ttl, name="simple", template=_key_template)
                return cached
            result = await func(*args, **kwargs)
            if condition(result, args, kwargs, key=_cache_key):
                await backend.set(_cache_key, result, expire=_ttl)
            return result

        return _wrap

    return _decor
